/*
 * Copyright (c) 2013 Claus Assmann
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the license/LICENSE.3C file which can be found at the
 * top level of this source code distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: t-cert-0.c,v 1.13 2014/03/04 05:38:35 ca Exp $")
#include "sm/assert.h"
#include "sm/error.h"
#include "sm/io.h"
#include "sm/str.h"
#include "sm/tls.h"
#include "sm/tlscnf.h"
#include "sm/tlsmisc.h"
#include "sm/misc.h"
#include "sm/test.h"

static int Verbose;

#define TESTCERT	"altcert.pem"
#define TESTKEY		"altkey.pem"
#define TESTCACERT	"CAalt.pem"
#define TESTCERTP	"certs"

static int
testcert(const char *name, int show_nA)
{
#if MTA_USE_TLS
	int r;
	sm_ret_T ret;
	tls_cnf_T tls_cnf;
	X509 *cert;
	SSL *ssl;
	sm_str_P str;
	sm_str_P strs[MTA_ALTNAMES_MAX];
	tls_ctx_P tls_ctx;
	X509_NAME *subj, *issuer;

	ret = sm_tls_init_library(&tls_ctx, false);
	SM_TEST_E(sm_is_success(ret));

	sm_memzero(&tls_cnf, sizeof(tls_cnf));
	tls_cnf.tlscnf_certfile = TESTCERT;
	tls_cnf.tlscnf_keyfile = TESTKEY;
	tls_cnf.tlscnf_cacertpath = TESTCERTP;
	tls_cnf.tlscnf_cacertfile = TESTCACERT;
	ret = sm_tls_init(tls_ctx, TLS_I_CLT, false, &tls_cnf);
	SM_TEST_E(ret == SM_SUCCESS);

	r = 1;
	ssl = SSL_new(tls_ctx->tls_ssl_ctx);
	SM_TEST_E(ssl != NULL);
	cert = SSL_get_certificate(ssl);
	SM_TEST_E(cert != NULL);

	str = sm_str_new(NULL, 256, 1024);
	SM_TEST_E(str != NULL);
	for (r = 0; r < SM_ARRAY_SIZE(strs); r++)
		strs[r] = NULL;
	ret = sm_tls_get_altname(cert, strs, SM_ARRAY_SIZE(strs));
	for (r = 0; r < ret; r++)
		sm_io_fprintf(smioerr, "subjectAltName[%d]=%S\n", r, strs[r]);
	if (r >= 0 && name != NULL) {
		sm_str_clr(str);
		sm_str_scat0(str, name);
		ret = mta_best_match(strs, SM_ARRAY_SIZE(strs), str, false);
		if (ret >= 0 && ret < SM_ARRAY_SIZE(strs))
			sm_io_fprintf(smioerr, "subjectAltName=%S\n", strs[ret]);
	}

	sm_str_clr(str);
	subj = X509_get_subject_name(cert);
	(void) sm_tls_get_cn(subj, NID_commonName, &str);
	sm_io_fprintf(smioerr, "cn_subj=%S\n", str);

	sm_str_clr(str);
	issuer = X509_get_issuer_name(cert);
	(void) sm_tls_get_cn(issuer, NID_commonName, &str);
	sm_io_fprintf(smioerr, "cn_issuer=%S\n", str);

	if (show_nA > 0) {
		time_T notAfter;

		notAfter = notafter(cert);
		if (1 == show_nA)
			sm_io_fprintf(smioerr, "notafter=%v\n", notAfter);
		else
			sm_io_fprintf(smioerr, "notafter=%s", ctime(&notAfter));
	}

	return 0;

 error:
	return -1;
#else /* MTA_USE_TLS */
	return 0;	/* skipped */
#endif /* MTA_USE_TLS */
}

static void
usage(const char *prg)
{
	sm_io_fprintf(smioerr, "%s: usage [options] cert\n", prg);
	sm_io_fprintf(smioerr, "-e n   show notAfter (n=1: seconds, n=2: ctime)\n");
	sm_io_fprintf(smioerr, "-v     show compile time options\n");
	sm_io_fprintf(smioerr, "-V     increase verbosity\n");
}

int
main(int argc, char *argv[])
{
	int c;
	const char *name;
	int show_nA = false;

	while ((c = getopt(argc, argv, "e:vV")) != -1) {
		switch (c) {
		  case 'e':
			show_nA = strtoul(optarg, NULL, 0);
			break;
		  case 'v':
#if MTA_USE_TLS
			sm_io_fprintf(smioerr, "MTA_USE_TLS\n");
#endif
			break;
		  case 'V':
			++Verbose;
			break;
		  default:
			usage(argv[0]);
			return(1);
		}
	}
	if (argc > 1)
		name = argv[1];
	else
		name = NULL;
	sm_test_begin(argc, argv, "test cert (subjectAltName) 0");
	testcert(name, show_nA);
	return sm_test_end();
}

