/*
 * Copyright (c) 2004-2006 Sendmail, Inc. and its suppliers.
 *      All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: occ.c,v 1.62 2014/01/30 02:25:31 ca Exp $")
#include "sm/assert.h"
#include "sm/magic.h"
#include "sm/str.h"
#include "sm/memops.h"
#include "sm/pthread.h"
#include "sm/evthr.h"
#include "sm/qmgr-int.h"
#include "sm/map.h"
#include "occ.h"

/*
**  Note: locking occ_ctx for all operations can cause locking contentions.
**  In most cases it should be sufficient to lock just occ_entry.
**  "exception": the lookup strings (occ_ctx->occx_lhs etc) must be protected
**  (alternatively local strings can be used again with the possible drawback
**  that an allocation may fail).
*/

#if OCC_RATE
/*
**  OCC_RATE_EXCEEDED -- is connection rate limit exceeded?
**
**	Parameters:
**		occ_entry -- OCC entry
**		now -- current time
**
**	Returns:
**		true iff connection rate limit is exceeded
*/

bool
occ_rate_exceeded(occ_entry_P occ_entry, time_T now)
{
	int cnt;
	uint u;
	ulong ticks;
	bool exceeded;
	conntime_P ct;

	SM_IS_OCCE(occ_entry);
	if (occ_entry->occe_max_rate <= 0)
		return 0;

	cnt = 0;
	ticks = now / CHTGRAN;
	ct = &occ_entry->occe_times[ticks % CHTSIZE];
	if (ct->ct_ticks != ticks) {
		ct->ct_ticks = ticks;
		ct->ct_count = 0;
	}
	++ct->ct_count;
	occ_entry->occe_rate_first = 0;
	for (u = 0; u < CHTSIZE; ++u) {
		ct = &occ_entry->occe_times[u];
		if (ct->ct_ticks <= ticks && ct->ct_ticks >= ticks - CHTSIZE) {
			cnt += ct->ct_count;
			if (0 == occ_entry->occe_rate_first)
				occ_entry->occe_rate_first = ct->ct_ticks * CHTGRAN;
		}
	}
	occ_entry->occe_rate = (cnt * 60) / (CHTSIZE * CHTGRAN);
	QM_LEV_DPRINTFC(QDC_OCC, 2, (QM_DEBFP, "sev=INFO, func=%s, ip=%I, open_se=%u, open_ta=%u, flags=%x, rate=%d\n", SMFCT, &occ_entry->occe_srv_ip, occ_entry->occe_open_se, occ_entry->occe_open_ta, occ_entry->occe_flags, occ_entry->occe_rate));

	exceeded = occ_entry->occe_rate > occ_entry->occe_max_rate;
	if (exceeded)
		--occ_entry->occe_times[ticks % CHTSIZE].ct_count;

	return exceeded;
}
#endif

/*
**  OCC_SE_LI_MAP -- get session limits from qmgr_conf map
**
**	Parameters:
**		qmgr_ctx -- QMGR context (only qmgr_conf_map is needed)
**		occ_ctx -- OCC context
**		srv_ip -- IP address of server
**		occ_entry -- OCC entry
**
**	Returns:
**		SM_SUCCESS
*/

static sm_ret_T
occ_se_li_map(qmgr_ctx_P qmgr_ctx, occ_ctx_P occ_ctx, sm_ipaddr_P srv_ip, occ_entry_P occ_entry)
{
	if (NULL == qmgr_ctx->qmgr_conf_map)
		return SM_SUCCESS;

/* tags for conf db */
#define OCITAG "oci:"	/* outgoing connections initial */
#define OCMTAG "ocm:"	/* outgoing connections max */
#define OCRTAG "ocr:"	/* outgoing connections rate */
#define OCTOTAG "octo:"	/* outgoing connection cache timeout */

#define OCC_GET_VAL(tag, entry)	do {	\
		sm_str_clr(occ_ctx->occx_tag);	\
		sm_str_clr(occ_ctx->occx_rhs);	\
		if (sm_str_scat(occ_ctx->occx_tag, tag) == SM_SUCCESS	\
		    && sm_map_lookup_ip(qmgr_ctx->qmgr_conf_map,	\
		                        occ_ctx->occx_lhs, occ_ctx->occx_tag,	\
		                        SMMAP_LFL_SUBNETS|SMMAP_LFL_TAG,	\
		                        occ_ctx->occx_rhs) == SM_SUCCESS	\
		    && sm_str_getlen(occ_ctx->occx_rhs) > 0)	\
		{				\
			ulong v;	\
						\
			errno = 0;	\
			v = strtoul((char *)sm_str_getdata(occ_ctx->occx_rhs), NULL, 0);\
			if (v != ULONG_MAX && errno != ERANGE)	\
				entry = (uint) v;	\
		}	\
	} while (0)

	sm_str_clr(occ_ctx->occx_lhs);
	sm_ipaddr2str(srv_ip, occ_ctx->occx_lhs);

	OCC_GET_VAL(OCITAG, occ_entry->occe_cur_conc = occ_entry->occe_init_conc);
	OCC_GET_VAL(OCMTAG, occ_entry->occe_max_conc);
#if OCC_RATE
	OCC_GET_VAL(OCRTAG, occ_entry->occe_max_rate);
#endif
	OCC_GET_VAL(OCTOTAG, occ_entry->occe_timeout);

	QM_LEV_DPRINTFC(QDC_OCC, 4, (QM_DEBFP, "sev=INFO, func=%s, ip=%I, conc_init=%u, conc_max=%u, tmo=%d\n", SMFCT, srv_ip, occ_entry->occe_cur_conc, occ_entry->occe_max_conc, occ_entry->occe_timeout));
	return SM_SUCCESS;
}


/* configuration based on rcpt domain IP */
struct scclc_S
{
	char	*scclc_name;
};
typedef struct scclc_S scclc_T, *scclc_P;

static sm_conf_definition_T
scclc_defs[] = {
{ SM_CONF_DEF_MAGIC, "connection_limits",		sm_conf_type_string,
	offsetof(scclc_T, scclc_name),	0,
	NULL, 0, NULL, NULL,	NULL,
	"connection limits"
	SM_LC_NO_ISSET	SM_LC_SET_MAGIC(0) },

CONF_SENTINEL
};

/*
**  OCC_GET_CL -- get name of connection limits from configuration
**
**	Parameters:
**		conf_str -- configuration data
**		cl -- name of connection limits (output)
**
**	Returns:
**		usual return code
*/

static sm_ret_T
occ_get_cl(sm_str_P conf_str, sm_str_P cl)
{
	int err;
	sm_conf_T *cnf;
	scclc_T scclc;

	SM_REQUIRE(conf_str != NULL);
	SM_REQUIRE(cl != NULL);

	cnf = sm_conf_new(SMFCT);
	if (NULL == cnf) {
		err = errno;
		return sm_error_temp(SM_EM_SMTPC, ENOMEM);
	}
	err = sm_conf_read_data(cnf, (char *)sm_str_getdata(conf_str),
				sm_str_getlen(conf_str), true);
	if (err != 0) {
		/* sm_prt_conferr(SMFCT, cnf, err, smioerr); */
		goto error;
	}

	err = sm_conf_scan(cnf, scclc_defs, 0, &scclc);
	if (err != 0) {
		/* sm_prt_conferr("map", cnf, err, smioerr); */
		goto error;
	}

	if (NULL == scclc.scclc_name) {
		err = SM_NOMATCH;
		goto error;
	}

	sm_str_clr(cl);
	err = sm_str_scat(cl, scclc.scclc_name);
	SM_CONF_DESTROY(cnf);
	return err;

  error:
	SM_CONF_DESTROY(cnf);
	return err;
}

/*
**  OCC_GET_LIMITS -- find connection limits
**
**	Parameters:
**		qmgr_ctx -- QMGR context
**		cl -- name of connection limits
**
**	Returns:
**		>=0: index in connection_limits array
**		<0: usual return code
*/

static sm_ret_T
occ_get_limits(qmgr_ctx_P qmgr_ctx, sm_str_P cl)
{
	uint u, n;
	const uchar *cl_name;
	size_t cl_len;

	n = qmgr_ctx->qmgr_cnf.q_cnf_sc_cls.sccls_n;
	cl_name = sm_str_getdata(cl);
	cl_len = sm_str_getlen(cl);
	for (u = 0; u < n; u++) {
		char *name;

		name = qmgr_ctx->qmgr_cnf.q_cnf_sc_cls.sccls_cl[u].sccl_name;
		if (cl_len != strlen(name))
			continue;
		if (!sm_memeq(name, cl_name, cl_len))
			continue;
		return u;
	}
	return sm_error_perm(SM_EM_OCC, SM_E_NOTFOUND);
}

/*
**  OCC_SE_LIMIT -- get session limits from conf
**
**	Parameters:
**		qmgr_ctx -- QMGR context
**		aq_rcpt -- one recipient for a transaction
**		conf -- string to store conf
**		tag -- string to store tag
**
**	Returns:
**		>=0: index in conf array
**		<0: usual error code
*/

sm_ret_T
occ_se_limit(qmgr_ctx_P qmgr_ctx, aq_rcpt_P aq_rcpt, sm_str_P conf, sm_str_P tag)
{
	sm_ret_T ret;

	sm_str_clr(conf);
	sm_str_clr(tag);

	QM_LEV_DPRINTFC(QDC_OCC, 4, (QM_DEBFP, "func=%s, rcpt_domain=%#S\n", SMFCT, aq_rcpt->aqr_domain));

#define SC_SE_RCPTDOM_TAG "smtpc_conf_rcpt_domain:"
	ret = sm_str_scat(tag, SC_SE_RCPTDOM_TAG);
	if (sm_is_err(ret))
		return ret;
	ret = sm_map_lookup_domain(qmgr_ctx->qmgr_conf_map,
		(sm_rdstr_P) aq_rcpt->aqr_domain, tag, 0,
		/* SMMAP_LFL_DOT | SMMAP_LFL_DOTSUBDOM */
		conf);
	if (sm_is_success(ret) && sm_str_getlen(conf) > 0) {
		ret = occ_get_cl(conf, tag);
		if (ret != SM_SUCCESS)
			return ret;
		ret = occ_get_limits(qmgr_ctx, tag);
		return ret;
	}

	sm_str_clr(conf);
	sm_str_clr(tag);
	return ret;
}

/*
**  OCC_SE_LI_CONF -- get session limits from conf
**
**	Parameters:
**		qmgr_ctx -- QMGR context
**		occ_ctx -- OCC context
**		aq_rcpt -- one recipient for a transaction
**		occ_entry -- OCC entry
**
**	Returns:
**		SM_SUCCESS
*/

static sm_ret_T
occ_se_li_conf(qmgr_ctx_P qmgr_ctx, occ_ctx_P occ_ctx, aq_rcpt_P aq_rcpt, occ_entry_P occ_entry)
{
	sm_ret_T ret;
	sc_cl_P sc_cl;
	uint v;

	if (NULL == qmgr_ctx->qmgr_conf_map ||
		!QCNF_IS_FLAG(qmgr_ctx, QCNF_FL_SC_LKP_CL_CONF))
	{
		return SM_SUCCESS;
	}

	if (NULL == occ_entry->occe_conf)
		occ_entry->occe_conf = sm_str_new(NULL, 64, 256); /* check size */
	if (NULL == occ_entry->occe_tag)
		occ_entry->occe_tag = sm_str_new(NULL, 32, 64); /* check size */
	if (NULL == occ_entry->occe_conf || NULL == occ_entry->occe_tag)
		return sm_error_temp(SM_EM_OCC, ENOMEM);

	ret = occ_se_limit(qmgr_ctx, aq_rcpt, occ_entry->occe_conf, occ_entry->occe_tag);
	if (sm_is_err(ret))
		return ret;	/* convert? */

	sc_cl = &(qmgr_ctx->qmgr_cnf.q_cnf_sc_cls.sccls_cl[ret]);

	v = sc_cl->sccl_init_conc_conn;
	if (v > 0)
		occ_entry->occe_cur_conc = occ_entry->occe_init_conc = v;
	v = sc_cl->sccl_max_conc_conn;
	if (v > 0)
		occ_entry->occe_max_conc = v;
#if OCC_RATE
	v = sc_cl->sccl_max_msg_rate;
	if (v > 0)
		occ_entry->occe_max_rate = v;
#endif
	return SM_SUCCESS;
}

/*
**  OCC_SESS_OPEN -- open a session in OCC
**
**	Parameters:
**		qmgr_ctx -- QMGR context
**		occ_ctx -- OCC context
**		srv_ip -- IP address of server
**		aq_rcpt -- one recipient for a transaction
**		pocc_entry -- pointer to OCC entry (output)
**		locktype -- kind of locking
**
**	Returns:
**		usual sm_error code; ENOMEM (see occ_entry_new())
**
**	Side Effects: none on error (except if unlock fails)
**
**	Locking: locks occ_ctx during operation if requested, returns unlocked
**
**	Last code review: 2005-03-17 02:23:27
**	Last code change:
*/

sm_ret_T
occ_sess_open(qmgr_ctx_P qmgr_ctx, occ_ctx_P occ_ctx, sm_ipaddr_P srv_ip, aq_rcpt_P aq_rcpt, occ_entry_P *pocc_entry, thr_lock_T locktype
#if MTA_USE_TLS && FFR_CERT_PINNING
	, tlscp_ctx_P certfpinfo
#endif
	)
{
	sm_ret_T ret;
	int r;
	occ_entry_P occ_entry;
	time_T now;

	now = evthr_time(qmgr_ctx->qmgr_ev_ctx);
	if (thr_lock_it(locktype)) {
		r = smthread_mutex_lock(&occ_ctx->occx_mutex);
		SM_LOCK_OK(r);
	}

	/* find entry for IP address, if it doesn't exist: create new one */
	ret = occ_entry_find(occ_ctx->occx_ht, srv_ip, &occ_entry,
			&occ_ctx->occx_mutex, THR_NO_LOCK);
	if (sm_is_err(ret)) {
		ret = occ_entry_new(occ_ctx->occx_ht, &occ_ctx->occx_fl_hd, srv_ip,
				&occ_entry, &occ_ctx->occx_mutex, THR_NO_LOCK);
		if (sm_is_err(ret))
			goto error;
		SM_IS_OCCE(occ_entry);
		occ_entry->occe_init_conc = qmgr_ctx->qmgr_cnf.q_cnf_init_conc_conn;
		occ_entry->occe_cur_conc = qmgr_ctx->qmgr_cnf.q_cnf_init_conc_conn;
		occ_entry->occe_max_conc = qmgr_ctx->qmgr_cnf.q_cnf_max_conc_conn;

		(void) occ_se_li_conf(qmgr_ctx, occ_ctx, aq_rcpt, occ_entry);

		/* for now: IP addresses override rcpt domain */
		(void) occ_se_li_map(qmgr_ctx, occ_ctx, srv_ip, occ_entry);

	}
	else {
		SM_ASSERT(occ_entry != NULL);
		SM_ASSERT(IPADDR_EQ(occ_entry->occe_srv_ip, *srv_ip));
	}

#if MTA_USE_TLS && FFR_CERT_PINNING
/* TODO: use q_cnf_max_delay, not occe_timeout */
	if (0 != occ_entry->occe_certfpinfo.tlscp_hs_stat &&
	    occ_entry->occe_last_upd + occ_entry->occe_timeout <= now)
	    occ_entry->occe_certfpinfo.tlscp_hs_stat = 0;
	cpcertfpinfo(certfpinfo, &occ_entry->occe_certfpinfo);
#endif

	/* increment counters etc */
	++occ_entry->occe_open_se;
	++occ_entry->occe_open_ta;
#if 0
	occ_entry->occe_last_conn = now;
#endif
	occ_entry->occe_last_upd = now;
	QM_LEV_DPRINTFC(QDC_OCC, 4, (QM_DEBFP, "sev=INFO, func=%s, occ_entry=%p, ip=%I, open_se=%u, open_ta=%u, flags=%x\n", SMFCT, occ_entry, &occ_entry->occe_srv_ip, occ_entry->occe_open_se, occ_entry->occe_open_ta, occ_entry->occe_flags));
	if (pocc_entry != NULL)
		*pocc_entry = occ_entry;
	if (thr_unl_no_err(locktype)) {
		r = smthread_mutex_unlock(&occ_ctx->occx_mutex);
		SM_UNLOCK_OK(r);
	}
	return SM_SUCCESS;

  error:
	if (thr_unl_if_err(locktype)) {
		r = smthread_mutex_unlock(&occ_ctx->occx_mutex);
		SM_UNLOCK_OK(r);
	}
	return ret;
}

/*
**  OCC_SESS_CLOSE_ENTRY -- close a session in OCC
**
**	Parameters:
**		occ_ctx -- OCC context
**		srv_ip -- IP address of server
**		ok -- session was ok?
**		now -- current time
**		pflags -- flags of session (output, may be NULL)
**		locktype -- kind of locking
**
**	Returns:
**		usual sm_error code; SM_E_NOTFOUND, (un)lock errors
**
**	Side Effects: none on error (except if unlock fails)
**
**	Locking: locks occ_ctx if requested
**
**	Last code review: 2005-03-14 18:54:53
**	Last code change:
*/

sm_ret_T
occ_sess_close_entry(occ_ctx_P occ_ctx, sm_ipaddr_P srv_ip, bool ok, time_T now, uint32_t *pflags, thr_lock_T locktype, dadb_entry_P dadb_entry)
{
	sm_ret_T ret;
	int r;
	uint32_t flags;
	occ_entry_P occ_entry;

	SM_IS_OCCX(occ_ctx);
	ret = SM_SUCCESS;
	flags = 0;
	if (thr_lock_it(locktype)) {
		r = smthread_mutex_lock(&occ_ctx->occx_mutex);
		SM_LOCK_OK(r);
	}

	/* access locked via occ_ctx */
	occ_entry = NULL;
	ret = occ_entry_find(occ_ctx->occx_ht, srv_ip, &occ_entry,
			&occ_ctx->occx_mutex, THR_NO_LOCK);
	if (SM_SUCCESS == ret) {
		SM_IS_OCCE(occ_entry);
		if (occ_entry->occe_open_se == 0) {
			QM_LEV_DPRINTFC(QDC_OCC, 0, (QM_DEBFP, "sev=ERROR, func=%s, ip=%I, open_se=0\n", SMFCT, &occ_entry->occe_srv_ip));
		}
		else {
			QM_LEV_DPRINTFC(QDC_OCC, 3, (QM_DEBFP, "sev=DBG, func=%s, ip=%I, ok=%d, open_se=%u, cur_conc=%u, flags=%x\n", SMFCT, &occ_entry->occe_srv_ip, ok, occ_entry->occe_open_se, occ_entry->occe_cur_conc, occ_entry->occe_flags));

#if MTA_USE_TLS && FFR_CERT_PINNING
			cpcertfpinfo(&occ_entry->occe_certfpinfo, &dadb_entry->dadbe_certfpinfo);
#endif
			if (ok && occ_entry->occe_open_se >= occ_entry->occe_cur_conc
				&& occ_entry->occe_cur_conc < occ_entry->occe_max_conc)
				++occ_entry->occe_cur_conc;
			else if (!ok && occ_entry->occe_cur_conc > 0)
				--occ_entry->occe_cur_conc;

			--occ_entry->occe_open_se;
			if (occ_entry->occe_open_se < occ_entry->occe_cur_conc) {
				flags |= OCCE_FL_BLW_LIM;
				if (OCCE_IS_FLAG(occ_entry, OCCE_FL_SE_WAIT)) {
					flags |= OCCE_FL_SE_WAIT;
					OCCE_CLR_FLAG(occ_entry, OCCE_FL_SE_WAIT);
				}
			}
		}
		if (occ_entry->occe_open_ta == 0) {
			QM_LEV_DPRINTFC(QDC_OCC, 3, (QM_DEBFP, "sev=ERROR, func=%s, ip=%I, open_ta=0\n", SMFCT, &occ_entry->occe_srv_ip));
		}
		else
			occ_entry->occe_open_ta--;
		if (occ_entry->occe_open_se == 0) {
#if !DA_OCC_RSC
			bht_rm(occ_ctx->occx_ht, HT_IPA_KEYaLEN(occ_entry->occe_srv_ip),
				NULL, NULL);
			occ_entry_free(&occ_ctx->occx_fl_hd, occ_entry,
					&occ_ctx->occx_mutex, THR_NO_LOCK);
#endif /* !DA_OCC_RSC */
		}
	}
	else {
		QM_LEV_DPRINTFC(QDC_OCC, 4, (QM_DEBFP, "sev=ERROR, func=%s, ip=%I, occ_entry_find=%x\n", SMFCT, &occ_entry->occe_srv_ip, ret));
	}

	if ((!sm_is_err(ret) && thr_unl_no_err(locktype))
	    || (sm_is_err(ret) && thr_unl_if_err(locktype)))
	{
		r = smthread_mutex_unlock(&occ_ctx->occx_mutex);
		SM_UNLOCK_OK(r);
	}
	if (pflags != NULL)
		*pflags = flags;
	return ret;
}

/*
**  OCC_SESS_REUSE -- reuse a session in OCC: increase number of open TAs
**
**	Parameters:
**		occ_ctx -- OCC context
**		occ_entry -- OCC entry
**		now -- current time
**		locktype -- kind of locking
**
**	Returns:
**		SM_SUCCESS except for (un)lock errors
**
**	Side Effects: none on error (except if unlock fails)
**
**	Locking: locks occ_ctx if requested
**
**	Last code review: 2005-03-14 18:56:05
**	Last code change:
*/

sm_ret_T
occ_sess_reuse(occ_ctx_P occ_ctx, occ_entry_P occ_entry, time_T now, thr_lock_T locktype)
{
	int r;

	SM_IS_OCCX(occ_ctx);
	SM_IS_OCCE(occ_entry);
	if (thr_lock_it(locktype)) {
		r = smthread_mutex_lock(&occ_ctx->occx_mutex);
		SM_LOCK_OK(r);
	}

	++occ_entry->occe_open_ta;

	if (thr_unl_no_err(locktype)) {
		r = smthread_mutex_unlock(&occ_ctx->occx_mutex);
		SM_UNLOCK_OK(r);
	}
	return SM_SUCCESS;
}

/*
**  OCC_TA_CLOSE_ENTRY -- close a transaction in OCC
**
**	Parameters:
**		occ_ctx -- OCC context
**		srv_ip -- IP address of server
**		now -- current time
**		pflags -- flags of session (output, may be NULL)
**		locktype -- kind of locking
**
**	Returns:
**		usual sm_error code; SM_E_NOTFOUND, (un)lock errors
**
**	Side Effects: none on error (except if unlock fails)
**
**	Locking: locks occ_ctx if requested
**
**	Last code review: 2005-03-14 18:56:53
**	Last code change:
*/

sm_ret_T
occ_ta_close_entry(occ_ctx_P occ_ctx, sm_ipaddr_P srv_ip, time_T now, uint32_t *pflags, thr_lock_T locktype)
{
	int r;
	sm_ret_T ret;
	uint32_t flags;
	occ_entry_P occ_entry;

	SM_IS_OCCX(occ_ctx);
	flags = 0;
	if (thr_lock_it(locktype)) {
		r = smthread_mutex_lock(&occ_ctx->occx_mutex);
		SM_LOCK_OK(r);
	}

	occ_entry = NULL;
	ret = occ_entry_find(occ_ctx->occx_ht, srv_ip,
			&occ_entry, &occ_ctx->occx_mutex, THR_NO_LOCK);
	if (SM_SUCCESS == ret) {
		SM_IS_OCCE(occ_entry);
		if (occ_entry->occe_open_ta == 0)
			QM_LEV_DPRINTFC(QDC_OCC, 4, (QM_DEBFP, "sev=ERROR, func=%s, entry=%p, ip=%I, open_ta=%d, state=before_decrement\n", SMFCT, occ_entry, &occ_entry->occe_srv_ip, occ_entry->occe_open_ta));
		else {
			occ_entry->occe_open_ta--;
			if (OCCE_IS_FLAG(occ_entry, OCCE_FL_TA_WAIT)
			    && occ_entry->occe_open_ta < occ_entry->occe_cur_conc)
			{
				flags |= OCCE_FL_TA_WAIT;
				OCCE_CLR_FLAG(occ_entry, OCCE_FL_TA_WAIT);
			}
		}

		QM_LEV_DPRINTFC(QDC_OCC, 4, (QM_DEBFP, "sev=INFO, func=%s, occ_entry=%p, ip=%I, open_ta=%d, state=after_decrement, occ_entry=%p, occe_flags=%x, flags=%x\n", SMFCT, occ_entry, &occ_entry->occe_srv_ip, occ_entry->occe_open_ta, occ_entry, occ_entry->occe_flags, flags));
	}
	else {
		QM_LEV_DPRINTFC(QDC_OCC, 4, (QM_DEBFP, "sev=ERROR, func=%s, ip=%I, occ_entry_find=%x\n", SMFCT, &occ_entry->occe_srv_ip, ret));
	}

	if (thr_unl_no_err(locktype)) {
		r = smthread_mutex_unlock(&occ_ctx->occx_mutex);
		SM_UNLOCK_OK(r);
	}
	if (pflags != NULL)
		*pflags = flags;
	return SM_SUCCESS;
}

/*
**  OCC_CLOSE -- close an OCC
**
**	Parameters:
**		occ_ctx -- OCC context
**
**	Returns:
**		SM_SUCCESS
**
**	Side Effects: free occ
**
**	Locking: no locking, destroys OCC (and hence lock)
**
**	Last code review: 2005-03-16 05:33:13
**	Last code change:
*/

sm_ret_T
occ_close(occ_ctx_P occ_ctx)
{
	if (NULL == occ_ctx)
		return SM_SUCCESS;
	SM_IS_OCCX(occ_ctx);
#if DA_OCC_RSC
	rsc_free(occ_ctx->occx_ht);
#else
	bht_destroy(occ_ctx->occx_ht, NULL, NULL);
#endif
	pthread_mutex_destroy(&occ_ctx->occx_mutex);
	SM_STR_FREE(occ_ctx->occx_lhs);
	SM_STR_FREE(occ_ctx->occx_tag);
	SM_STR_FREE(occ_ctx->occx_rhs);
#if OCC_CHECK
	occ_ctx->sm_magic = SM_MAGIC_NULL;
#endif
	sm_free_size(occ_ctx, sizeof(*occ_ctx));
	return SM_SUCCESS;
}

#if DA_OCC_RSC
/*
**  RSC_OCCE_CREATE -- "create" an OCC entry (callback function for rsc)
**
**	Parameters:
**		key -- ignored
**		len -- ignored
**		value -- OCC entry
**		ctx -- ignored
**
**	Returns:
**		value
**
**	Last code review: 2005-03-17 04:50:31
**	Last code change:
*/

static void *
rsc_occe_create(const char *key, uint len, void *value, void *ctx)
{
	return value;
}

/*
**  RSC_OCCE_DELETE -- "delete" an OCC entry (callback function for rsc)
**
**	Parameters:
**		value -- OCC entry
**		ctx -- OCC context
**
**	Returns:
**		SM_SUCCESS (see occ_entry_free())
**
**	Last code review: 2005-03-16 05:38:32
**	Last code change:
*/

static sm_ret_T
rsc_occe_delete(void *value, void *ctx)
{
	occ_ctx_P occ_ctx;

	occ_ctx = (occ_ctx_P) ctx;
	SM_IS_OCCX(occ_ctx);
	return occ_entry_free(&occ_ctx->occx_fl_hd, (occ_entry_P) value,
			&occ_ctx->occx_mutex, THR_NO_LOCK);
}
#endif /* DA_OCC_RSC */

/*
**  OCC_OPEN -- open a new OCC context
**
**	Parameters:
**		pocc_ctx -- pointer to OCC context (output)
**		size -- size of OCC
**
**	Returns:
**		usual sm_error code; ENOMEM,
**
**	Side Effects: none on error
**
**	Last code review: 2005-03-16 05:42:34
**	Last code change:
*/

sm_ret_T
occ_open(occ_ctx_P *pocc_ctx, uint size)
{
	int r;
	sm_ret_T ret;
	occ_ctx_P occ_ctx;

/* these are upper limits which are a bit larger than necessary */
#define SM_CONF_STR_MAX		64	/* max length of IP address + tag */
#define SM_CONF_IP		40	/* length of IP address */
#define SM_CONF_TAG		8	/* length of tag */
#define SM_CONF_RHS		48	/* length of IP address + tag */

	SM_REQUIRE(pocc_ctx != NULL);
	SM_REQUIRE(size > 0);
	occ_ctx = (occ_ctx_P) sm_zalloc(sizeof(*occ_ctx));
	if (NULL == occ_ctx)
		return sm_error_temp(SM_EM_OCC, ENOMEM);

	/* use some prime number?? */
#if DA_OCC_RSC
	occ_ctx->occx_ht = rsc_new(size * 2 + 1, size * 3 + 3, 0,
		rsc_occe_create, rsc_occe_delete, (void *) occ_ctx);
#else
	occ_ctx->occx_ht = bht_new(size, size * 2 + 1);
#endif
	if (NULL == occ_ctx->occx_ht)
		goto enomem;
	OCCFL_INIT(&occ_ctx->occx_fl_hd);
	occ_ctx->occx_lhs = sm_str_new(NULL, SM_CONF_IP, SM_CONF_STR_MAX);
	if (NULL == occ_ctx->occx_lhs)
		goto enomem;
	occ_ctx->occx_tag = sm_str_new(NULL, SM_CONF_TAG, SM_CONF_STR_MAX);
	if (NULL == occ_ctx->occx_tag)
		goto enomem;
	occ_ctx->occx_rhs = sm_str_new(NULL, SM_CONF_RHS, SM_CONF_STR_MAX);
	if (NULL == occ_ctx->occx_rhs)
		goto enomem;

	r = pthread_mutex_init(&occ_ctx->occx_mutex, SM_PTHREAD_MUTEXATTR);
	if (r != 0) {
		ret = sm_error_perm(SM_EM_OCC, r);
		goto error;
	}
#if 0
//	occ_ctx->occ_entries_max = size;
//	occ_ctx->occ_entries_lim = size;
#endif

#if OCC_CHECK
	occ_ctx->sm_magic = SM_OCCX_MAGIC;
#endif
	*pocc_ctx = occ_ctx;
	return SM_SUCCESS;

  enomem:
	ret = sm_error_temp(SM_EM_OCC, ENOMEM);
  error:
	if (occ_ctx != NULL)	/* just paranoia */ {
		if (occ_ctx->occx_ht != NULL) {
#if DA_OCC_RSC
			rsc_free(occ_ctx->occx_ht);
#else
			bht_destroy(occ_ctx->occx_ht, NULL, NULL);
#endif
		}
		SM_STR_FREE(occ_ctx->occx_lhs);
		SM_STR_FREE(occ_ctx->occx_tag);
		SM_STR_FREE(occ_ctx->occx_rhs);
		sm_free_size(occ_ctx, sizeof(*occ_ctx));
	}
	return ret;
}
