/*
 * Copyright (c) 2012 Claus Assmann
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the license/LICENSE.3C file which can be found at the
 * top level of this source code distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: logentry.c,v 1.65 2014/01/30 16:37:29 ca Exp $")
#include "sm/assert.h"
#include "sm/error.h"
#include "sm/ctype.h"
#include "sm/io.h"
#include "sm/log.h"
#include "sm/syslog.h"
#include "sm/base64.h"
#include "sm/str.h"
#include "sm/str-int.h"
#if !SM_NO_CSTR
#include "sm/cstr.h"
#endif
#if MTA_USE_PTHREADS
#include "sm/pthread.h"
#endif
#include "log.h"
#define COND_PRINT 1
#include "print.h"
#include "sm/net.h"

#include <stdio.h>

struct logt_S {
	const char	*lt_name;
	uint		 lt_spec;
	uint32_t	 lt_dont;
	/* should we have width, prec here? */
};
typedef struct logt_S logt_T, *logt_P;

static logt_T logtypes[] = {
	 { "",			SLT(SLT_END, SM_LOGT_OFF),	LTD_NONE	}	/* dummy, not used */
	,{ NULL,		SLT(SLT_FREE, 's'),		LTD_NONE	}
	,{ "sev",		SLT(SLT_SEV, 's'),		LTD_NONE	}
	,{ "func",		SLT(SLT_FUNC, 's'),		LTD_FUNC	}
	,{ "where",		SLT(SLT_WHERE, 's'),	LTD_WHERE	}
	,{ "ss_se",		SLT(SLT_SS_SE, 's'),	LTD_NONE	}
	,{ "ss_ta",		SLT(SLT_SS_TA, 's'),	LTD_NONE	}
	,{ "mail",		SLT(SLT_MAIL, 'N')|SLT_CLEAN,	LTD_NONE	}
	/* ? */
	,{ "se_id",		SLT(SLT_SE_ID, 's'),	LTD_NONE	}
	,{ "ss_id",		SLT(SLT_SS_ID, 'd'),	LTD_NONE	}
	,{ "ta_id",		SLT(SLT_TA_ID, 's'),	LTD_NONE	}
	,{ "ta",		SLT(SLT_TA, 's'),	LTD_NONE	}
	/* next two are the same? */
	,{ "rcpt",		SLT(SLT_RCPT, 'T')|SLT_CLEAN,	LTD_NONE	}	/* %@N, @S, @T, p */
	,{ "rcpt_pa",	SLT(SLT_RCPT_PA, 'S'),	LTD_NONE	}	/* %S, .256S */
	,{ "rcpt_id",	SLT(SLT_RCPT_ID, 's'),	LTD_NONE	}
	,{ "rcpt_idx",	SLT(SLT_RCPT_IDX, 'u'),	LTD_NONE	}
	,{ "rcpt_bounce",	SLT(SLT_RCPT_B, 's'),	LTD_NONE	}
	/* aren't the next two the same? */
	,{ "rcpt_bounce_idx",	SLT(SLT_RCPT_B_IDX, 'u'),	LTD_NONE	}	/* %d, u */
	,{ "rcpt_idx_dsn",	SLT(SLT_RCPT_IDX_DSN, 'u'),	LTD_NONE	}	/* %d, u */
	,{ "nrcpts",	SLT(SLT_NRCPTS, 'u'),	LTD_NONE	}
	,{ "i",			SLT(SLT_I, 'u'),	LTD_NONE	}
	,{ "cdb",		SLT(SLT_CDB, 'C'),		LTD_NONE	}
	,{ "msgid",		SLT(SLT_MSGID, 'N')|SLT_HASH,	LTD_NONE	}	/* %#N */
	,{ "size",		SLT(SLT_SIZE, 'u')|SLT_L,	LTD_NONE	}	/* %lu */
	,{ "stat",		SLT(SLT_STAT, 'd'),		LTD_NONE	}
	,{ "stat",		SLT(SLT_STAT_M, 'm'),		LTD_NONE	}
	,{ "state",		SLT(SLT_STATE, 'd'),		LTD_NONE	}
	,{ "status",	SLT(SLT_STATUS, 's'),	LTD_NONE	}
	,{ "status",	SLT(SLT_STATUS_T, 'T')|SLT_CLEAN,	LTD_NONE	}
	,{ "status",	SLT(SLT_STATM, 'm'),	LTD_NONE	}
	,{ "error",		SLT(SLT_ERROR, 's'),	LTD_NONE	}
	,{ "input",		SLT(SLT_INPUT, 'T')|SLT_CLEAN,	LTD_NONE	}
	,{ "module",	SLT(SLT_MOD, 'H'),		LTD_NONE	}
	,{ "da_se",		SLT(SLT_DA_SE, 's'),	LTD_NONE	}
	,{ "da_ta",		SLT(SLT_DA_TA, 's'),	LTD_NONE	}
	,{ "reuse",		SLT(SLT_REUSE, 'd'),	LTD_NONE	}
	,{ "keep_open",		SLT(SLT_KEEP_OPEN, 'd'),	LTD_NONE	}
	,{ "entries_cur",	SLT(SLT_E_CUR, 'd'),	LTD_NONE	}
	,{ "entries_max",	SLT(SLT_E_MAX, 'd'),	LTD_NONE	}
	,{ "open_connections",		SLT(SLT_OPN_CONN, 'd'),	LTD_NONE	}
	,{ "cur_limit",		SLT(SLT_CUR_LIMIT, 'u'),	LTD_NONE	}
	,{ "cur_rate",		SLT(SLT_CUR_RATE, 'd'),	LTD_NONE	}
	,{ "max_rate",		SLT(SLT_RATE_MAX, 'u'),	LTD_NONE	}
	,{ "timeout",		SLT(SLT_TIMEOUT, 'd'),	LTD_NONE	}
	,{ "cert_issuer",	SLT(SLT_CERT_ISS, 'S')|SLT_HASH,	LTD_NONE	}
	,{ "cert_subject",	SLT(SLT_CERT_SUB, 'S')|SLT_HASH,	LTD_NONE	}
	,{ "common_name",	SLT(SLT_CN, 'S')|SLT_HASH,			LTD_NONE	}
	,{ "cipher",		SLT(SLT_CIPHER, 'S'),				LTD_NONE	}
	,{ "cipher_bits",	SLT(SLT_USE_BITS, 'd'),			LTD_NONE	}
	,{ "alg_bits",		SLT(SLT_ALG_BITS, 'd'),			LTD_NONE	}
	,{ "se_reused",		SLT(SLT_TLS_REUSED, 'w'),		LTD_NONE	}
	,{ "verify",		SLT(SLT_VERIFY, 's'),		LTD_NONE	}
	,{ "client_ip",		SLT(SLT_CLT, 'a'),				LTD_NONE	}
	,{ "client_ip",		SLT(SLT_CLT_IP, 'S'),			LTD_NONE	}
	,{ "client_name",	SLT(SLT_CLT_NAME, 'C'),			LTD_NONE	}
	,{ "server",		SLT(SLT_SRV, 'a'),				LTD_NONE	}
	,{ "reply",			SLT(SLT_REPLY, 'S')|SLT_CLEAN,	LTD_NONE	}
	,{ "ip",		SLT(SLT_IP, SM_LOGT_OFF),	LTD_NONE }	/* %A, I, a, #S */
	,{ "ipv4",		SLT(SLT_IPV4, 'A'),		LTD_NONE	}
	,{ "ipv6",		SLT(SLT_IPV6, 'B'),		LTD_NONE	}
	,{ "port",		SLT(SLT_PORT, 'd'),		LTD_NONE	}	/* %d, hd */
	,{ "lock",		SLT(SLT_LOCK, 'd'),		LTD_NONE	}
	,{ "ntries",	SLT(SLT_NTRIES, 'd'),	LTD_NONE	}
	,{ "mt_RHS",	SLT(SLT_MT_RHS, SM_LOGT_OFF),	LTD_NONE	}	/* \"%@.256S\" */
	,{ "thread",	SLT(SLT_SC_THR, 'u'),	LTD_NONE	}
	,{ "pid",		SLT(SLT_PID, 'd'),	LTD_NONE	}
	,{ "max_cur_thrs",	SLT(SLT_SS_MCT, 'u'),	LTD_NONE	}
	,{ "Avail",		SLT(SLT_SS_AVLB, 'd'),	LTD_NONE	}
	/* many different meanings? */
	,{ "idx",		SLT(SLT_IDX_U, 'u'),	LTD_NONE	}
	,{ "idx",		SLT(SLT_IDX_D, 'd'),	LTD_NONE	}
	,{ "ret",		SLT(SLT_RET, 'd'),		LTD_NONE	}
	/* need to implement "modifiers"! */
	,{ "delay",		SLT(SLT_DELAY, 'u')|SLT_L,	LTD_NONE	}	/* %lu */
	,{ "xdelay",	SLT(SLT_XDELAY, 'u')|SLT_L,	LTD_NONE	}	/* %lu */
	,{ "t_connect",	SLT(SLT_CDELAY, 'u'),	LTD_NONE	}	/* %u */
	,{ "t_ehlo",	SLT(SLT_EDELAY, 'u'),	LTD_NONE	}	/* %u */
	,{ "t_starttls",SLT(SLT_TDELAY, 'u'),	LTD_NONE	}	/* %u */
	,{ "t_mail",	SLT(SLT_MDELAY, 'u'),	LTD_NONE	}	/* %u */
	,{ "t_rcpts",	SLT(SLT_RDELAY, 'u'),	LTD_NONE	}	/* %u */
	,{ "t_data",	SLT(SLT_DDELAY, 'u'),	LTD_NONE	}	/* %u */
	,{ "t_msg",		SLT(SLT_BDELAY, 'u'),	LTD_NONE	}	/* %u */
	,{ "t_eom",		SLT(SLT_FDELAY, 'u'),	LTD_NONE	}	/* %u */
	,{ "now-entered",	SLT(SLT_NOW_ENTERED, 'd')|SLT_L,	LTD_NONE	}	/* %lu */
	,{ "iid",		SLT(SLT_IID, 's'),			LTD_NONE	}
	,{ "pri",		SLT(SLT_PRI, 'd'),			LTD_NONE	}
	,{ "dkim_signed",	SLT(SLT_DKIMS, 's'),	LTD_NONE	}
	,{ NULL,		SLT(SLT_LAST, SM_LOGT_OFF),	LTD_NONE	}
/*
	,{ "",		SLT(SLT_, 'd'),	LTD_NONE	}
*/
};

#if SM_LOG_TEST
sm_ret_T
sm_letest(void)
{
	uint u;

	if (SLT_LAST != SM_ARRAY_SIZE(logtypes))
		return -1;
	for (u = 0; u <= SLT_LAST; u++) {
		if (u != sm_log_field(logtypes[u].lt_spec))
			return u + 1;
	}
	return SM_SUCCESS;
}
#endif /* SM_LOG_TEST */

/*
**  SM_LOG_ENTRY -- Write a logfile entry
**
**	Parameters:
**		lctx -- log context
**		category -- category
**		module -- module
**		priority -- priority
**		level -- log level
**		... -- arguments
**
**	Returns:
**		SM_SUCCESS except for (un)lock errors and wrong calls
*/

sm_ret_T
sm_log_entry(sm_log_ctx_P lctx, sm_logcategory_P category, sm_logmodule_P module, int priority, uint level, ...)
{
	uint32_t led, lt;
	uint ui, l;
	int wrt2str, ac;
#define LOG_W2S_NO		0
#define LOG_W2S_YES		1
#define LOG_W2S_FLUSH	2
	bool first, inva, printit, prttm;
	sm_ret_T ret;
	logt_T logtype;
	char fc;
	const char *name;
	sm_file_T *fp;
	sm_file_T fake;

/* copy from vprintf.c */
	int flags;             /* flags as above */
	int width;             /* width from format (%8d), or 0 */
	int prec;              /* precision from format (%.3d), or -1 */
	char *encode;          /* characters to be encoded */
	int dprec;             /* a copy of prec if [diouxX], 0 otherwise */
	int realsz;            /* field size expanded by dprec */
	int size;              /* size of converted field or string */
	int osize;             /* original size before sanitizing */
	char sign;             /* sign prefix (' ', '+', '-', or \0) */
	char ox[2];            /* space for 0x hex-prefix */
	int ch;                /* character from fmt */
	int n, m, n2;          /* handy integers (short term usage) */
	ulonglong_T _uquad;    /* integer arguments %[diouxX] */
	const char *cp;        /* handy const char pointer (short term usage) */
	char *wcp;             /* handy char pointer (short term usage) */
	char *xdigs="0123456789abcdef"; /* digits for [xX] conversion */
	char buf[BUF];         /* space for %c, %[diouxX], %[eEfgG] */
	sm_str_P str;
#if !SM_NO_CSTR
	sm_cstr_P cstr;
#endif
	enum { OCT, DEC, HEX, BASE32, BASE64 } base;/* base for [diouxX] conversion */
#define SM_B32DIGS "0123456789ABCDEFGHIJKLMNOPQRSTUV"
	const char *b64digs = SM_B64DIGS;
	const char *b32digs = SM_B32DIGS;

	/*
	**  Choose PADSIZE to trade efficiency vs. size.  If larger printf
	**  fields occur frequently, increase PADSIZE and make the initialisers
	**  below longer.
	*/
#define PADSIZE	16		/* pad chunk size */
	static char blanks[PADSIZE] =
	 {' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' '};
	static char zeroes[PADSIZE] =
	 {'0','0','0','0','0','0','0','0','0','0','0','0','0','0','0','0'};

	/* see vfprintf.c and io-int.h */
	struct sm_iov *iovp;   /* for VFP_PRINT macro */
#define NIOV 8
	sm_uio_T uio;		/* output information: summary */
	sm_iov_T iov[NIOV];	/* ... and individual io vectors */
	va_list ap;
/* end copy from vprintf.c */

#define GETARG(type) va_arg(ap, type)

/*
if lctx is NULL: create a string, add a file on top of it, and write to that?
we also need a string (lctx->lctx_str) to invoke syslog()
*/

	inva = false;
	if (lctx != NULL) {
		SM_IS_LOGCTX(lctx);
		if (!sm_log_wouldlog(lctx, category, module, level))
			return SM_SUCCESS;
	}
	else
		return sm_error_perm(SM_EM_LOG, EINVAL);

	SM_REQUIRE(lctx != NULL);
	ret = SM_SUCCESS;
	LOCK_LOG;

#define SETUP_FP(str) do {	\
		/* While snprintf(3) specifies size_t stdio uses an int internally */ \
		l = sm_str_getlen(str);	\
		ui = sm_str_getmax(str) - l;	\
		if (ui > INT_MAX)	\
			ui = INT_MAX;	\
		/* question: put this into a static?? */	\
		fake.sm_magic = SM_FILE_MAGIC;	\
		f_fd(fake) = -1;	\
		f_flags(fake) = SMWR|SMSTRSTR;	\
		fake.f_cookie = (void *)(str);	\
		f_bfbase(fake) = f_p(fake) = (str)->sm_str_base + l;	\
		f_bfsize(fake) = f_w(fake) = ui ? ui - 1 : 0;	\
		fake.f_timeout = SM_TIME_FOREVER;	\
		f_read(fake) = NULL;	\
		f_write(fake) = NULL;	\
		f_close(fake) = NULL;	\
		f_open(fake) = NULL;	\
		f_seek(fake) = NULL;	\
		f_setinfo(fake) = f_getinfo(fake) = NULL;	\
		fp = &fake;	\
	} while (0)

	/*
	 *  Should we do that in the setup code?
	 *  That will probably make a lot of code in log.c simpler.
	 */

	if (NULL == lctx->lctx_fp) {
		sm_str_clr(lctx->lctx_str);
		SETUP_FP(lctx->lctx_str);
	}
	else
		fp = lctx->lctx_fp;

	uio.uio_iov = iovp = iov;
	uio.uio_resid = 0;
	uio.uio_iovcnt = 0;

	va_start(ap, level);
	inva = true;
	first = true;
	wrt2str = LOG_W2S_NO;
	prttm = fp == lctx->lctx_fp; /* write timestamp only for a local file */
	for (ac = 0;; ac++) {

		flags = 0;
		dprec = 0;
		width = 0;
		prec = -1;
		sign = '\0';

		/* avoid "might be used uninitialized" */
		size = 0;
		osize = 0;
		encode = "";
		cp = "";
		ch = '\0';

		led = GETARG(uint32_t);
		if (SLT_END == led)
			break;
		if (0 == ac && SLT_IS_SET_FP(led)) {
			str = GETARG(sm_str_P);
			if (NULL == str)
				/* complain? abort? */
				break;
			SETUP_FP(str);
			if (prttm && SLT_SET_FP_FIRST == led)
				sm_io_fputs(fp, (const uchar *) sm_log_tstamp(LOG_MICRO(lctx)));
			prttm = false;
			if (SLT_SET_FP_FIRST != led)
				first = false;
			wrt2str = (SLT_SET_FP_LAST == led) ? LOG_W2S_FLUSH : LOG_W2S_YES;
			continue;
		}
		if (prttm) {	
			sm_io_fputs(lctx->lctx_fp, (const uchar *) sm_log_tstamp(LOG_MICRO(lctx)));
			prttm = false;
		}
		l = sm_log_field(led);
		if (l > SM_ARRAY_SIZE(logtypes)) {
			/* complain? abort? */
			break;
		}
		logtype = logtypes[l];

		/* paranoia: make sure table and macros do not get out of sync */
		SM_ASSERT(l == sm_log_field(logtype.lt_spec));
		lt = sm_log_type(led);
		if (lt != 0)
			fc = sm_log_t2f(led);
		else
			fc = sm_log_t2f(logtype.lt_spec);
		ch = fc;

		if (sm_log_iscond(led))
			printit = GETARG(int);
		else
			printit = 0 == (lctx->lctx_dont & logtype.lt_dont);

		/* there is supposed to be another entry */
		if (!first)
			VFP_PRINT(", ", 2);
#if SM_LOG_SEVERITY
		else
			/* this must be done differently! use VFP_PRINT() */
			sm_io_fprintf(fp, "sev=%s, ", sm_log_sev2txt(priority));
#endif

		n = (lt != 0) ? sm_log_mod(led) : sm_log_mod(logtype.lt_spec);
		switch (n) {
		  case SLT_LV:
			flags |= LONGINT;
			break;
		  case SLT_LLV:
			flags |= LLONGINT;
			break;
		  case SLT_SV:
			flags |= SHORTINT;
			break;
		  case SLT_CV:
			flags |= CHARINT;
			break;
		  default:
			break;
		}

		if (sm_log_ishash(led) || (lt == 0 && sm_log_ishash(logtype.lt_spec)))
			flags |= ALT;
		else if (sm_log_isxtext(led) || (lt == 0 && sm_log_isxtext(logtype.lt_spec))) {
			flags |= ALT;
			encode = "+, <>\"\\()";
		}
		else if (sm_log_isclean(led) ||
			(lt == 0 && sm_log_isclean(logtype.lt_spec)))
		{
			flags |= ALT;
			encode = "%,";
		}
		if (sm_log_iswidth(led))
			width = GETARG(int);
		if (sm_log_isprec(led))
			prec = GETARG(int);
		if (sm_log_is0pad(led))
			flags |= ZEROPAD;

		if (SLT_FREE == l) {
			name = GETARG(char *);
			if (printit && NULL == name) {
				/* complain? abort? */
				ret = sm_error_perm(SM_EM_LOG, EINVAL);
				break;
			}
			VFP_PRINT(name, strlen(name));
			VFP_PRINT("=", 1);
		}
		else if ((name = logtype.lt_name) != NULL) {
			if (SLT_STATM != l) {
				VFP_PRINT(name, strlen(name));
				VFP_PRINT("=", 1);
			}
		}

		switch (fc) {
#include "print.c"

		  default:
			ret = sm_error_perm(SM_EM_LOG, EINVAL);
			goto error;
		}

#include "printit.c"
		if (printit)
			first = false;
	}
	va_end(ap);
	VFP_FLUSH();

	if (LOG_W2S_FLUSH == wrt2str) {
		str = (sm_str_P)fake.f_cookie;
		if (NULL == lctx->lctx_fp) {
			syslog(priority, "%.*s", (int) sm_str_getlen(str),
				(const char *) sm_str_data(str));
		}
		else {
			sm_io_fputstr(lctx->lctx_fp, str);
			sm_putc(lctx->lctx_fp, '\n');
		}
	}
	else if (LOG_W2S_NO == wrt2str) {
		if (NULL == lctx->lctx_fp) {
			syslog(priority, "%.*s", (int) sm_str_getlen(lctx->lctx_str),
				(const char *) sm_str_data(lctx->lctx_str));
		}
		else
			sm_putc(lctx->lctx_fp, '\n');
	}

#if SM_LOG_ROTATE
	if (++lctx->lctx_entries > LOG_ENTRIES_MAX)
		sm_log_rotate(lctx);
#endif
	UNLOCK_LOG;
	return ret;

 error:
	if (inva)
		va_end(ap);
	UNLOCK_LOG;
	return ret;
}
